from multiprocessing.pool import ThreadPool
from typing import Callable, List, Any, Iterable


def parallel_map(fn: Callable[[Any], Any], inputs: Iterable[Any], num_thread: int):
    """Applies a function to a list of inputs in parallel.

    Uses a Threading pool to process inputs in parallel.

    Args:
        fn: A `Callable` function that takes an element from `inputs` as argument.
        inputs: A `Iterable` object.
        num_thread: An `int` scalar represents the number of threads.

    Returns:
        A corresponding list of processed inputs.
    """
    if num_thread <= 1:
        return list(map(fn, inputs))
    pool = ThreadPool(num_thread)
    results = pool.map(fn, inputs)
    pool.close()
    pool.join()
    return results


def try_get_attr(instance, attr_name):
    if hasattr(instance, attr_name):
        return getattr(instance, attr_name)
    else:
        return None


def get_state_dict(obj, exclusions=None,recursive=True):
    keys = vars(obj).keys()
    state_dict = {}

    for key in keys:
        if not exclusions or key not in exclusions:
            prop = getattr(obj, key)
            if recursive:
                to_state_dict = try_get_attr(prop, 'state_dict')
                if callable(to_state_dict):
                    prop = to_state_dict()
            state_dict[key] = prop

    return state_dict


def load_state_dict(obj, state_dict):
    for key, val in state_dict.items():
        prop = getattr(obj, key)
        load_state_dict = try_get_attr(prop, 'load_state_dict')
        if callable(load_state_dict):
            load_state_dict(val)
        else:
            setattr(obj, key, val)



from typing import List, Dict, Callable, Any

from overrides import overrides

from lunas.batch import Batch, Cache
from lunas.interface import Persistable
from lunas.reader import Reader
from lunas.utils import get_state_dict, load_state_dict


class DataIterator(Persistable):
    """An iterator that iterates through a `Reader`.

    This class performs multi-pass iterations over the dataset and maintains
    the iteration state.
    """

    def __init__(self, reader: Reader, batch_size, cache_size: int = 1000, sample_size_fn: Callable[[Any], int] = None,
                 collate_fn: Callable[[List[Any]], Any] = None, sort_desc_by: Callable[[Any], int] = None):
        """Initialize the iterator.

        Args:
            reader: A `Reader` object.
            batch_size: A `int` scalar that limits the size of returned batch.
            cache_size: A `int` scalar. Prefetch `cache_size` samples from the `reader` in `self.cache`.
            sample_size_fn: (Optional.) A callable function that calculates size for each sample.
                The size of each sample will then be summed up as the size of the batch. If not
                specified, default to 1 for each sample, which is actually `lambda sample: 1`.
            collate_fn: (Optional.) A callable function that converts a list of samples to model inputs.
            sort_desc_by: (Optional.) A callable function that returns a sorting key for each sample. If not
                specified, leave the batch as it is.
        """
        super().__init__()
        self.reader = reader
        self.batch_size = batch_size
        self.sample_size_fn = sample_size_fn
        self.collate_fn = collate_fn
        self.cache_size = cache_size
        self.sort_desc_by = sort_desc_by

        if sample_size_fn is None:
            sample_size_fn = lambda sample: 1
        if collate_fn is None:
            collate_fn = lambda samples: None

        self.sample_size_fn = sample_size_fn
        self.collate_fn = collate_fn

        # bookkeeping params
        self.step_in_epoch = -1
        self.step = -1
        self.epoch = 0

        self.cache = Cache(cache_size, sample_size_fn)
        self.remains = []

        self._exclusions = ['sample_size_fn', 'collate_fn', 'sort_desc_by']

        self.check_batch_size(batch_size, cache_size)
        self.reset()

    def set_batch_size(self, batch_size) -> None:
        """Allows dynamic batch size at runtime.

        Args:
            batch_size: A `int` scalar.

        """
        self.check_batch_size(batch_size)
        self.batch_size = batch_size

    def check_batch_size(self, batch_size, cache_size=None) -> None:
        """Checks whether batch_size is < cache_size.

        To ensure rationality, batch_size must be < cache_size.

        Args:
            batch_size: A `int` scalar.
            cache_size: A `int` scalar.

        """
        cache_size = cache_size or self.cache_size
        if batch_size > cache_size:
            raise RuntimeError(f'Batch size should be less than cache size. '
                               f'Got batch_size = {batch_size} and cache_size = {cache_size}')

    def reset(self):
        self.step_in_epoch = -1
        self.step = -1
        self.epoch = 0
        self.remains.clear()
        self.cache.pop_all()
        self.reader = iter(self.reader)

    def iter_epoch(self):
        """Iterate through the dataset for one epoch.

        For the last batch, it will be dropped if its size is smaller
        than 2/3 of the specified batch size.

        """

        cache = self.cache
        remains = self.remains

        end_of_epoch = False
        sort_batch = False
        while True:
            batch = Batch(self.batch_size, self.sample_size_fn)
            if cache.effective_size() < self.batch_size * 2 / 3.0:
                if end_of_epoch:
                    self.reader = iter(self.reader)
                    break
                # Consume samples from cache before filling-in
                remains += cache.pop_all()
                try:
                    # Fill cache
                    cache.from_iter(self.reader, raise_when_stopped=True)
                except StopIteration:
                    # Mark as end
                    end_of_epoch = True
                cache.sort(self.sort_desc_by)

            # if self.epoch==2 and not xxx:
            #     print(cache._samples)
            #     xxx=True
            if remains:
                batch.from_list(remains, self.batch_size)
                sort_batch = True
            batch.from_iter(cache, self.batch_size)

            if batch.effective_size() < self.batch_size:
                remains += batch.pop_all()
            else:
                if sort_batch:
                    batch.sort(self.sort_desc_by)
                    sort_batch = False
                # yield (batch, self.collate_fn(batch.samples))

                self.step_in_epoch += 1
                self.step += 1
                yield (batch, None)
        raise StopIteration

    def while_true(self, predicate: Callable[[], bool]):
        """Iterates through the dataset by a given stopping criteria.

        Args:
            predicate: A callable function. This function is evaluated to determine
            whether iteration should continue or not.

        Returns:
            (batch, inputs): A `Tuple` consists of a `Batch` object and model inputs. When `self.collate_fn`
                is None, the returned `inputs` is also None.
        """
        epoch_iter = self.iter_epoch()

        while predicate():
            try:
                batch = next(epoch_iter)
            except StopIteration:
                self.epoch += 1
                epoch_iter = self.iter_epoch()
                continue

            yield batch

    def __call__(self, while_predicate: Callable[[], bool]):
        return self.while_true(while_predicate)

    @overrides
    def state_dict(self) -> Dict:
        return get_state_dict(self, exclusions=self._exclusions, recursive=True)

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        load_state_dict(self, state_dict)



import abc
from typing import List, Iterable, Dict, Callable, Any

import numpy
from overrides import overrides

from lunas.interface import Persistable
from lunas.utils import parallel_map, get_state_dict, load_state_dict


class BaseReader(Iterable, Persistable):
    """A abstract class representing a dataset reader.

    A reader includes pipeline to process each sample and iterate through the dataset.
    This class defines any interface that's visible to users.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super().__init__()
        # Excludes these attributes from `self.state_dict()`
        self._exclusions: List[str] = []

    def cursor(self) -> int:
        """Increases the cursor and returns the new value.

        Returns:
            A `int` indicates current position.
        """
        raise NotImplementedError

    def reset_cursor(self) -> None:
        """Reset cursor to enable re-iterating over the dataset.

        """

    def where(self, predicate: Callable[[Any], bool]):
        """Filters a sample by predicate.

        A predicate is a callable object that takes a sample as input and returns a
        `bool` value to indicate whether this sample should be filtered out.

        Note that the predicate is not applied to the dataset immediately. Instead, we
        apply predicates to a large buffer and utilizes multi-threading pool to speed up data
        pipeline.

        Args:
            predicate: A `Callable` object.

        Returns:
            `self`.
        """
        raise NotImplementedError

    def select(self, fn: Callable[[Any], Any]):
        """Transforms a sample.

        A fn is a callable object that takes a sample as input and returns a transformed
        sample.

        Note that the transformation is not applied to the dataset immediately. Instead, we
        apply transformations to a large buffer and utilizes multi-threading pool to speed up data
        pipeline.

        Args:
            fn: A `Callable` object.

        Returns:
            `self`.

        """
        raise NotImplementedError

    def size(self) -> int:
        """Get size of this dataset.

        Note: The returned value represents the total number of samples of the dataset, without filters
        applied to samples. When filters are applied, the effective size of this dataset may be different
        the value returned by this method. So please make sure the filters are applied to the final dataset
        wrapper to avoid size inconsistency.

        For example, the following usage of `self.where()` filters would cause size inconsistency.
        ```python
        ds1 = RangeReader(10).where(lambda x: x<5)
        ds2 = RangeReader(10).where(lambda x: x<6)
        ds = ZipReader(ds1, ds2)
        iterator = DataIterator(ds)
        # ...
        for batch in iterator():
            print(batch)
        ```

        The effective size of `ds1` at runtime is 5, while it's 6 for `ds2`. This inconsistency cannot be checked
        before iterating through `ds`.

        Returns:
            A scalar of type `int`.

        """
        raise NotImplementedError

    def next(self) -> Any:
        """Get an original sample from the dataset by index. DO NOT invoke this method from
        outside, use `next(dataset)` instead.

        Returns:
            A sample of any type.
        """
        raise NotImplementedError

    def _next(self) -> Any:
        """Wrapper method to get next sample.

        Wraps `self.next()`, apply transformations and predicates, and maintains
        internal state of this object.

        Returns:
            A sample of any type.

        """
        raise NotImplementedError

    def _apply(self, sample: Any) -> Any:
        """Applies transformations and filters to a sample.

        When sample is filtered out, returns None, else returns transformed sample.

        Args:
            sample: An input sample to be processed.

        Returns:
            A sample or `None` if it's filtered.
        """
        raise NotImplementedError

    def finalize(self):
        """Release resources after iteration stops.

        """
        raise NotImplementedError

    def __len__(self):
        """Returns size of this dataset.

        Returns:
            A `int` scalar.
        """
        return self.size()

    def __next__(self) -> Any:
        """Implements iterator method and returns a sample.

        Returns:
            A sample instance.
        """
        raise NotImplementedError

    def __iter__(self):
        """Returns an iterator of this dataset.

        """
        self.reset_cursor()
        return self

    def shuffle(self, buffer_size: int = 10000, num_threads: int = 1):
        """Returns a shufflable counterpart of this dataset.

        Args:
            buffer_size: A `int` scalar. Represents the number of samples to shuffle each time.
            num_threads:

        Returns:
            A `ShuffleReader`.

        """
        raise NotImplementedError

    def load_state_dict(self, state_dict: Dict) -> None:
        raise NotImplementedError

    def state_dict(self):
        raise NotImplementedError


class Reader(BaseReader):
    """Implements the functions defined in `BaseReader`.

    """

    def __init__(self, buffer_size: int = 10000, num_threads: int = 1):
        super().__init__()
        self._buffer_size = buffer_size
        self._num_threads = num_threads
        self._buffer: List[Any] = []
        self._fns = []
        self._fast_skip: bool = False
        # Indicates position in the dataset.
        self._cursor: int = -1

        self._exclusions += ['_fns', '_buffer']

    @overrides
    def cursor(self) -> int:
        self._cursor += 1
        return self._cursor

    @overrides
    def reset_cursor(self):
        self._cursor: int = -1

    def check_buffer_size(self) -> None:
        """Evaluates the buffer size lazily to prevent ``self.size()`` function from invoking before
        sub-class is properly configured.

        """
        buffer_size = self._buffer_size
        buffer_size = buffer_size if buffer_size > 0 else len(self)
        self._buffer_size = min(buffer_size, len(self))

    def get_effective_buffer_size(self) -> int:
        return len(self._buffer)

    def _fill_buffer(self) -> int:
        self.check_buffer_size()
        buffer = self._buffer
        while self.get_effective_buffer_size() < self._buffer_size:
            try:
                buffer.append(self.next())
            except StopIteration:
                break
        size = len(buffer)
        if not self._fast_skip and size > 0:
            self._buffer = self._parallel_apply(self._buffer)
        return size

    @overrides
    def select(self, fn: Callable[[Any], Any]) -> BaseReader:
        self._fns.append(fn)
        return self

    @overrides
    def where(self, predicate: Callable[[Any], bool]) -> BaseReader:
        self._fns.append(predicate)
        return self

    @overrides
    def _apply(self, sample: Any) -> Any:
        for fn in self._fns:
            new_sample = fn(sample)
            # Stops when predicate is evaluated to False.
            if new_sample is False:
                return None
            elif new_sample is not True:
                sample = new_sample

        return sample

    def _parallel_apply(self, samples: List[Any]) -> List[Any]:
        return parallel_map(self._apply, samples, self._num_threads)

    @overrides
    def _next(self) -> Any:
        self.cursor()

        if self.get_effective_buffer_size() == 0:
            self._fill_buffer()
        try:
            return self._buffer.pop(0)
        except IndexError:
            raise StopIteration

    @overrides
    def state_dict(self) -> Dict:
        return get_state_dict(self, exclusions=self._exclusions)

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        load_state_dict(self, state_dict)
        cursor = self._cursor
        self.reset_cursor()
        # Fast-skip these samples
        self._fast_skip = True
        while self._cursor < cursor:
            self._next()
        self._buffer = self._parallel_apply(self._buffer)
        self._fast_skip = False

    @overrides
    def finalize(self):
        pass

    @overrides
    def __next__(self) -> Any:
        # Check whether reaches the end of the dataset.
        if self._cursor + 1 >= self.size():
            self.finalize()
            raise StopIteration

        sample = self._next()
        # Don't return filtered samples.
        if sample is None:
            return self.__next__()
        return sample

    def __iter__(self) -> BaseReader:
        return super().__iter__()

    def shuffle(self, buffer_size: int = 10000, num_threads: int = 1) -> BaseReader:
        return ShuffleReader(self, buffer_size, num_threads)


class ShuffleReader(Reader):
    def __init__(self, reader: BaseReader, buffer_size: int = 10000, num_threads: int = 1):
        super().__init__(buffer_size, num_threads)
        self.reader = reader
        self._random_state = numpy.random.get_state()

    @overrides
    def size(self) -> int:
        return len(self.reader)

    def next(self) -> Any:
        rv = self.reader.next()
        # print(rv)
        return rv

    def _shuffle_buffer(self):
        numpy.random.shuffle(self._buffer)

    @overrides
    def _fill_buffer(self):
        rv = super()._fill_buffer()
        self._shuffle_buffer()
        return rv

    def finalize(self):
        self.reader.finalize()
        super().finalize()

    @overrides
    def __iter__(self) -> Iterable:
        self.reader = iter(self.reader)
        return super().__iter__()

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        numpy.random.set_state(state_dict['_random_state'])
        del state_dict['_random_state']
        super().load_state_dict(state_dict)


class ZipReader(Reader):
    def __init__(self, *readers: List[BaseReader], buffer_size: int = 10000, num_threads: int = 1):
        super().__init__(buffer_size, num_threads)
        self.readers = readers
        lens = list(map(len, readers))
        if len(set(lens)) != 1:
            raise RuntimeError(f'Sizes of datasets must match. Got {lens}')
        self._exclusions += ['readers']

    @overrides
    def size(self) -> int:
        return len(self.readers[0])

    @overrides
    def next(self):
        sample = tuple([r.next() for r in self.readers])
        # sample = tuple(r.next() for r in self.readers)
        if any(s is None for s in sample):
            sample = None

        return sample

    def finalize(self):
        for reader in self.readers:
            reader.finalize()
        super().finalize()

    @overrides
    def __iter__(self) -> Iterable:
        self.readers = list(map(iter, self.readers))
        return super().__iter__()

    @overrides
    def state_dict(self) -> Dict:
        state = super().state_dict()
        state['readers'] = [r.state_dict() for r in self.readers]
        return state

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        for r, state in zip(self.readers, state_dict['readers']):
            r.load_state_dict(state)
        del state_dict['readers']

        super().load_state_dict(state_dict)


class RangeReader(Reader):
    def __init__(self, start: int, stop: int = None, step: int = None, buffer_size: int = 10000, num_threads: int = 1):
        super().__init__(buffer_size, num_threads)
        if stop is None:
            stop = start
            start = 0
        if step is None:
            step = 1

        self.start = start
        self.stop = stop
        self.step = step

        self._range = None

        self._exclusions += ['_range']

    @overrides
    def size(self) -> int:
        import math
        return int(math.ceil((self.stop - self.start) / self.step))

    def _reset(self):
        start, stop, step = self.start, self.stop, self.step
        self._range = iter(range(start, stop, step))

    @overrides
    def reset_cursor(self):
        super().reset_cursor()
        self._reset()

    @overrides
    def next(self) -> Any:
        return next(self._range)



from typing import List, Iterable, Dict, Callable, Any

import numpy
from overrides import overrides

from lunas.interface import Persistable
from lunas.utils import get_state_dict, load_state_dict


class Batch(Persistable):
    def __init__(self, size: int, size_fn: Callable[[Any], int]):
        super().__init__()
        self.size = size
        self.size_fn = size_fn
        self._samples = []
        self._sort_idx: numpy.ndarray = None
        self._sample_sizes: Dict[Any, int] = dict()
        self._exclusions = ['size_fn', ]

    @property
    def samples(self):
        return self._samples

    def pop_all(self):
        samples = self._samples
        self._samples = []
        self._sample_sizes = dict()
        return samples

    def pop(self, idx: int = -1):
        sample = self._samples.pop(idx)
        # Delete item only if there are no more identical samples.
        if sample not in self._samples:
            del self._sample_sizes[sample]
        return sample

    def push(self, sample):
        self._samples.append(sample)
        self._sample_sizes[sample] = self.size_fn(sample)

    def get_sample_sizes(self) -> Iterable:
        return (self._sample_sizes[x] for x in self._samples)

    def effective_size(self):
        return sum(self.get_sample_sizes())

    def from_iter(self, sample_iter: Iterable, size: int = None, raise_when_stopped: bool = False):
        """
        Fills batch from an iterable object. Enables dynamic batch size if size is not None.
        Args:
            sample_iter:
            size:
            raise_when_stopped:

        """
        size = size or self.size
        # Approximate
        while self.effective_size() < size:
            try:
                self.push(next(sample_iter))
            except StopIteration as e:
                if raise_when_stopped:
                    raise e
                else:
                    break

    def from_list(self, sample_list: List, size: int = None):
        size = size or self.size
        while self.effective_size() < size:
            try:
                self.push(sample_list.pop(0))
            except IndexError:
                break

    def strip_batch_by_size(self, size: int = None):
        """
        Ensure effective batch size is limited.
        Args:
            size:

        Returns:

        """
        size = size or self.size
        rv = []
        while self.effective_size() > size and len(self._samples) > 1:
            rv.append(self.pop(-1))
        return rv

    def sort(self, key_fn: Callable[[Any], int]):
        if key_fn is not None:
            keys = numpy.array(list(map(key_fn, self._samples)))
            indices = numpy.argsort(keys)[::-1]
            self._sort_idx = indices
            indices = list(indices)
            self._samples = [self._samples[i] for i in indices]

    def desort(self, samples: List[Any]):
        if len(samples) != len(self._sort_idx):
            raise RuntimeError(f'Number of samples must match the number of indices. '
                               f'{len(samples)} != {len(self._sort_idx)}')
        indices = numpy.argsort(self._sort_idx)
        indices = list(indices)
        samples = [samples[i] for i in indices]
        return samples

    @overrides
    def state_dict(self) -> Dict:
        return get_state_dict(self, exclusions=self._exclusions)

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        load_state_dict(self, state_dict)


class Cache(Batch, Iterable):
    def __init__(self, size: int, size_fn: Callable[[Any], int]):
        super().__init__(size, size_fn)

    def __next__(self):
        if self._samples:
            return self.pop(0)
        raise StopIteration

    def __iter__(self):
        return self



import abc
from typing import Dict


class Persistable(object):
    """An interface to introduce persistence method.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractclassmethod
    def state_dict(self) -> Dict:
        """Returns a dictionary representing the state of the class.

        The returned state dictionary should contain picklable attributes of `self`,
        excluding callable function object.

        Returns:
            A dictionary of `Dict` type represent, which represents the persistent state of `self`.

        """
        raise NotImplementedError

    @abc.abstractclassmethod
    def load_state_dict(self, state_dict: Dict) -> None:
        """Restores the state by a given state.

        Sets attributes of `self` by dictionary.

        Args:
            state_dict: A `Dict`.

        """
        raise NotImplementedError



import sys

from setuptools import setup, find_packages

if sys.version_info < (3,):
    sys.exit('Sorry, Python3 is required for thseq.')

with open('requirements.txt') as r:
    requires = [l.strip() for l in r]

setup(
    name='Lunas',
    version='0.1.6',
    author='Seann Zhang',
    author_email='pluiefox@live.com',
    packages=find_packages(),
    url='https://github.com/pluiez/lunas',
    license='LICENSE.txt',
    description='A data processing pipeline and iterator with minimal dependencies '
                'for machine learning.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',

)



from typing import TextIO, Any

from overrides import overrides

from lunas.reader import Reader


class TextReader(Reader):
    def __init__(self, filename: str, buffer_size: int = 10000, num_threads: int = 1):
        super().__init__(buffer_size, num_threads)
        self._filename = filename
        self._fd: TextIO = None
        self._exclusions += ['_fd']
        self._num_line=-1

    @overrides
    def size(self) -> int:
        if self._num_line<0:
            n = 0
            with open(self._filename) as r:
                for _ in r:
                    n += 1
            self._num_line = n
        return self._num_line

    @overrides
    def reset_cursor(self):
        self._reset()
        super().reset_cursor()

    def _reset(self):
        if self._fd is not None and not self._fd.closed:
            raise RuntimeError('Please close the file before reset.')
        self._fd = open(self._filename)

    @overrides
    def finalize(self):
        if self._fd is not None:
            self._fd.close()
        super().finalize()

    @overrides
    def next(self) -> Any:
        return self._fd.readline()
